using System;
using System.Collections;
using System.Reflection;

namespace Lazy
{
	/// <summary>
	/// Lazy object factory
	/// </summary>
	
	public class LazyFactory
	{
		/// <summary>
		/// Creates a lazy object of the given type
		/// </summary>
		/// <param name="type">A type (class or interface)</param>
		/// <returns>A new lazy object of the given type</returns>
		
		public static object Create (Type type) 
		{
			LazyObjectModel model;
			Type            lazyType;
			object          instance;

			instance = null;

			if (type.IsInterface) 
				type = GetClass(type);

			if (type!=null) {
				model = GetObjectModel(type);
				lazyType = GetLazyType(type,model);

				if (lazyType!=null) {
					instance = CreateInstance(lazyType);
				}
			}

			return instance;
		}

		/// <summary>
		/// LazyObjectModel cache
		/// </summary>

		private static Hashtable modelDictionary = new Hashtable();

		/// <summary>
		/// Gets the suitable lazy object model for a given type
		/// </summary>
		/// <param name="type">A type</param>
		/// <returns>The corresponding lazy object model</returns>

		public static LazyObjectModel GetObjectModel (Type type) {

			LazyObjectModel model = (LazyObjectModel) modelDictionary[type];

			if (model==null) {
				//Console.WriteLine("NEW MODEL");
				model = new LazyObjectModel(type);
				modelDictionary[type] = model;
			}

			return model;
		}

		/// <summary>
		/// LazyType cache
		/// </summary>

		private static Hashtable typeDictionary = new Hashtable();

		/// <summary>
		/// Gets the suitable lazy type for a given type
		/// </summary>
		/// <param name="type">A type</param>
		/// <returns>The corresponding lazy type</returns>

		public static Type GetLazyType (Type type, LazyObjectModel model) {

			Type lazy = (Type) typeDictionary[type];

			if (lazy==null) {
				//Console.WriteLine("NEW TYPE");
				lazy = LazyType.CreateType(model);
				typeDictionary[type] = lazy;
			}

			return lazy;
		}

		/// <summary>
		/// Creates an object of the given type (by calling its constructor without parameters)
		/// </summary>
		/// <param name="type">A type</param>
		/// <returns>A new object of the given type</returns>

		private static object CreateInstance (Type type) {

			object obj = null;
			ConstructorInfo constructor = type.GetConstructor
				( BindingFlags.Instance | BindingFlags.Public, null, 
				CallingConventions.HasThis, new Type[0], null );

			if (constructor!= null) {
				obj = constructor.Invoke(new object[0]);
			}
			
			return obj; // == Activator.CreateInstance(type);
		}

		// 

		/// <summary>
		/// Finds a class suitable for the given interface
		/// </summary>
		/// <param name="interfaz">An interface</param>
		/// <returns>A class implementing the interface</returns>

		private static Type GetClass (Type interfaz) {

			int        i;
			Assembly[] assemblies;
			Type       type = GetClassInAssembly(interfaz, interfaz.Assembly);

			if (type==null) {
				assemblies = AppDomain.CurrentDomain.GetAssemblies();

				for (i=0; i<assemblies.Length && (type==null); i++) {
					type = GetClassInAssembly(interfaz,assemblies[i]);
				}
			}

			return type;
		}

		/// <summary>
		/// Finds a class implementing the given interface in the given assembly
		/// </summary>
		/// <param name="interfaz">An interface</param>
		/// <param name="assembly">An assembly</param>
		/// <returns>A class implementing the interface included in the assembly</returns>

		private static Type GetClassInAssembly (Type interfaz, Assembly assembly) {

			Type[] types = assembly.GetTypes();

			foreach (Type type in types) {

				if (type.IsClass && interfaz.IsAssignableFrom(type)) {
					return type;
				}
			}

			return null;
		}
										   }
}
