using System;
using System.Reflection;
using System.Reflection.Emit;

namespace Lazy
{
	/// <summary>
	/// Descripcin breve de LazyMethod.
	/// </summary>
	public class LazyMethod
	{
		// Los atributos originales del mtodo deberan ser
		// MethodAttributes.Public | MethodAttributes.Virtual;

		public static void Define (TypeBuilder typeBuilder, LazyObjectModel model, int methodIndex, FieldInfo modelField) {
			MethodInfo method = (MethodInfo) model.GetLazyMethodImplementation(methodIndex,0).Method;

			String name = method.Name;
			Type returnType = method.ReturnType;
			Type[] parameters = GetParameterTypes(method);

			// Atributos del mtodo (VIRTUAL HIDEBYSIG !NEWSLOT)
			
			MethodAttributes attributes = method.Attributes | MethodAttributes.Virtual
				| MethodAttributes.HideBySig;
			attributes &= ~MethodAttributes.NewSlot;
							

			// Console.WriteLine("METHOD "+methodIndex+": "+name+" -> "+returnType+" ("+parameters.Length+")");
		
			MethodBuilder methodBuilder= typeBuilder.DefineMethod
				(name, attributes, returnType, parameters);

			ILGenerator methodIL = methodBuilder.GetILGenerator();

			Label labelElse = methodIL.DefineLabel();
			Label labelEnd  = methodIL.DefineLabel();

			// Variables locales
			// (loc.0): object[] parameters
			// (loc.1): int alternativa
			
			LocalBuilder localArray = methodIL.DeclareLocal(typeof(object[]));
			LocalBuilder localIndex = methodIL.DeclareLocal(typeof(int));

			// Alternativa adecuada
			// int alternative = model.LazyMethodIndex(this,3);  

			MethodInfo index = typeof(LazyObjectModel).GetMethod("LazyMethodIndex");

			methodIL.Emit(OpCodes.Ldarg_0);             // this
			methodIL.Emit(OpCodes.Ldfld, modelField);   // .model
			methodIL.Emit(OpCodes.Ldarg_0);             //   ( this,
			methodIL.Emit(OpCodes.Ldc_I4, methodIndex); //     methodIndex )
			methodIL.Emit(OpCodes.Callvirt, index);
			methodIL.Emit(OpCodes.Stloc_1);

			// if (alternative!=0) {
			methodIL.Emit(OpCodes.Ldloc_1);
			methodIL.Emit(OpCodes.Brfalse, labelElse);

			  // LazyObjectModel.InvokeMethod
			  DynamicCall(methodIL,parameters,returnType,methodIndex,modelField);

			methodIL.Emit(OpCodes.Br, labelEnd);
  
			// else
			methodIL.MarkLabel(labelElse);
			
			  // base.Method(...)
			  StaticCall(methodIL,name,parameters,model.Type);
			  
						
			// END

			methodIL.MarkLabel(labelEnd);

			// Devolucin del resultado

			methodIL.Emit(OpCodes.Ret);			
		}

		private static void StaticCall (ILGenerator methodIL, string methodName, Type[] parameters, Type baseType) 
		{
			methodIL.Emit(OpCodes.Ldarg_0); // this

			for (int i=0; i<parameters.Length; i++) {
				methodIL.Emit(OpCodes.Ldarg,i+1);
			}

			MethodInfo baseMethod = baseType.GetMethod(methodName,parameters);
			
			methodIL.Emit(OpCodes.Call, baseMethod);
		}


		private static void DynamicCall (ILGenerator methodIL, Type[] parameters, Type returnType, int methodIndex, FieldInfo modelField) 
		{
  			// Creacin del vector de parmetros

			methodIL.Emit(OpCodes.Ldc_I4, parameters.Length);
			methodIL.Emit(OpCodes.Newarr, typeof(Object));
			methodIL.Emit(OpCodes.Stloc_0);

			// Inicializacin del vector de parmetros (NB: boxing)

			for (int i=0; i<parameters.Length; i++) {
				methodIL.Emit(OpCodes.Ldloc_0);
				methodIL.Emit(OpCodes.Ldc_I4, i);
				methodIL.Emit(OpCodes.Ldarg, i+1);
				// Boxing (opcional)
				Box (methodIL, parameters[i]);
				// parameters[i]
				methodIL.Emit(OpCodes.Stelem_Ref);
  			}

			// model.InvokeMethodImplementation(this, methodIndex, alternative, object[]);
  
			methodIL.Emit(OpCodes.Ldarg_0);
			methodIL.Emit(OpCodes.Ldfld, modelField);   // model
			methodIL.Emit(OpCodes.Ldarg_0);             // this
			methodIL.Emit(OpCodes.Ldc_I4, methodIndex); // lazy method
			methodIL.Emit(OpCodes.Ldloc_1);             // alternative
			methodIL.Emit(OpCodes.Ldloc_0);             // parameters

			MethodInfo invokeMethod = typeof(LazyObjectModel).GetMethod("InvokeMethodImplementation");
			methodIL.Emit(OpCodes.Callvirt, invokeMethod);

			// Unboxing (opcional)

			Unbox(methodIL, returnType);
		}


		// Boxing

		private static void Box (ILGenerator methodIL, Type type) {

			if ((type!=null) && (type.IsValueType))
				methodIL.Emit(OpCodes.Box, type);
		}

		// Unboxing

		private static void Unbox (ILGenerator methodIL, Type type) 
		{
			if ((type!=null) && (type.IsValueType)) {
				methodIL.Emit(OpCodes.Unbox, type);
				methodIL.Emit(OpCodes.Ldobj, type);
			}
		}



		private static Type[] GetParameterTypes (MethodBase method) {

			ParameterInfo[] parameters = method.GetParameters();
			Type[]          parameterTypes = new Type[parameters.Length];

			for (int i=0; i<parameters.Length; i++)
				parameterTypes[i] = parameters[i].ParameterType;

			return parameterTypes;
		}
	}
}
