using System;
using System.Collections;
using System.Reflection;
using System.Reflection.Emit;

using Reflector.Disassembler;


namespace Lazy
{
	/// <summary>
	/// Descripcin breve de LazyMethodModel.
	/// </summary>
	public class LazyMethodModel
	{
		private LazyObjectModel model;
		private MethodBase method;     // MethodInfo || ConstructorInfo

		private ArrayList uses;     // Fields used
		private ArrayList defines;  // Fields defined
		private ArrayList calls;    // Methods called

		// Constructor

		public LazyMethodModel (LazyObjectModel model, MethodBase method, ModuleReader reader) {
			this.model = model;
			this.method = method;

			this.uses = new ArrayList();
			this.defines = new ArrayList();
			this.calls = new ArrayList();

			// Anlisis del cdigo intermedio

			ParseMethod(reader,method);
		}	

		private void ParseMethod (ModuleReader reader, MethodBase method) {

			MethodBody methodBody = reader.GetMethodBody(method);		
	
			
			foreach (Instruction instruction in methodBody.GetInstructions()) {

				OpCode code = instruction.Code;             // code.Name == OPCODE
				object operand = instruction.Operand;
								
				if ( operand!=null) {

					if ( code.OperandType == OperandType.InlineField ) {
						// FieldInfo operand (type == operand.DeclaringType)
						
						if ( code.Name.StartsWith("ld") ) {
							// Used fields (ldfld)
							Uses( (FieldInfo) operand );
						} else if (code.Name.StartsWith("st") ) {
							// Defined fields (stfld)
							Defines ( (FieldInfo) operand );
						}

					} else if (code.OperandType == OperandType.InlineMethod ) {
						// MethodBase operand (MethodInfo || ConstructorInfo)

						// Called methods (call|callvirt|newobj)
						Calls( (MethodBase) operand);
					}
				}
			}

		}

		private void Uses (FieldInfo field) {
			if (!uses.Contains(field))
				uses.Add(field);
		}

		private void Defines (FieldInfo field) {
			if (!defines.Contains(field))
				defines.Add(field);
		}

		private void Calls (MethodBase method) {
			if (!calls.Contains(method))
				calls.Add(method);
		}

		// uses = uses U uses(calls)
		// defines = defines U defines(calls)

		internal void ComputeDefUseClosure () {

			Queue     queue = new Queue(calls);
			ArrayList visited = new ArrayList();
			MethodBase method;
			LazyMethodModel methodModel;

			while (queue.Count>0) {

				method = (MethodBase) queue.Dequeue();

				if (!visited.Contains(method)) {

					if (method.DeclaringType == model.Type) {

						methodModel = model.GetMethod( MethodKey(method) );

						foreach (FieldInfo field in methodModel.uses)
							Uses(field);

						foreach (FieldInfo field in methodModel.defines)
							Defines(field);

						foreach (MethodInfo call in methodModel.calls)
							queue.Enqueue(call);
					}

					visited.Add(method);
				}
			}         
		}


		// Accessors

		public MethodBase Method {
			get { return method; }
		}

		public int UseCount () {
			return uses.Count;
		}

		// Salida estndar

		public override string ToString () {

			string str =  method.ToString();

			if (IsAlternativeImplementation()) 
				str += " [" +  AlternativeName() +"]";
			
			if (uses.Count>0)
  				str += "\n\tUSES"+ListFields(uses,"\n\t\t");

			if (defines.Count>0)
				str += "\n\tDEFINES"+ListFields(defines,"\n\t\t");

			if (calls.Count>0)
				str += "\n\tCALLS"+ListCalls(calls,"\n\t\t");

			return str;
		}

		private string ListFields (ArrayList array, string separator) {
			string str = "";

			foreach (FieldInfo field in array) {
				str += separator + field.DeclaringType + "::" + field.ToString();
			}

			return str;
		}


		private string ListCalls (ArrayList array, string separator) {
			string str = "";

			foreach (MethodBase method in array) {
				str += separator + method.DeclaringType + "::" + method.ToString();
			}

			return str;
		}
		
		// IsAlternativeImplementation

		public AlternativeImplementationAttribute Attribute () {
			object[] attributes = method.GetCustomAttributes(typeof(AlternativeImplementationAttribute),false);

			if (attributes.Length>0)
				return (AlternativeImplementationAttribute) attributes[0];
			else
				return null;
		}

		public bool IsAlternativeImplementation () {

			return Attribute()!=null;
		}

		public string AlternativeName () {
			return Attribute().Method;
		}

		public string AlternativeKey () {
			return MethodKey(AlternativeName(), method);
		}




		// USE/DEFINE/CALL lists

		public bool IsUsed (FieldInfo field) {
			return uses.Contains(field);
		}

		public bool IsDefined (FieldInfo field) {
			return defines.Contains(field);
		}

		public bool IsCalled (MethodBase method) {
			return calls.Contains(method);
		}

		// Check object

		public bool CheckObject (object obj) {

			bool      ok = model.Type.IsAssignableFrom(obj.GetType());
			FieldInfo field;

			for (int i=0; ok && (i<uses.Count); i++) {

				field = (FieldInfo) uses[i];

				if (!field.FieldType.IsPrimitive) {

					ok = (field.GetValue(obj) != null);

				} else if (typeof(long).IsAssignableFrom(field.FieldType)) {

					ok = ((long) field.GetValue(obj)) != 0;

				} else if (typeof(int).IsAssignableFrom(field.FieldType)) {

					ok = ((int) field.GetValue(obj)) != 0;
				
				} else if (typeof(double).IsAssignableFrom(field.FieldType)) {

					ok = ((double) field.GetValue(obj)) != 0;
					
				} else if (typeof(float).IsAssignableFrom(field.FieldType)) {

					ok = ((float) field.GetValue(obj)) != 0;


				} else if (typeof(bool).IsAssignableFrom(field.FieldType)) {

					ok = (bool) field.GetValue(obj);
				}
			}

			return ok;
		}

		// Method key

		public string MethodKey () {
			return MethodKey(method.Name,method);
		}

		public static string MethodKey (MethodBase method) {
			return MethodKey(method.Name,method);
		}

		private static string MethodKey (string name, MethodBase method) {

			ParameterInfo[] parameters = method.GetParameters();
			string          key = name;

			for (int i=0; i<parameters.Length; i++)
				key += ":"+parameters[i].ParameterType.ToString();
	
			return key;
		}

	}
}
