using System;
using System.Collections;
using System.Reflection;

using Reflector.Disassembler;


namespace Lazy
{
	/// <summary>
	/// Descripcin breve de LazyObjectModel.
	/// </summary>
	public class LazyObjectModel
	{
		private Type type;
		private ArrayList fields;  // Name -> LazyFieldModel
		private SortedList methods; // Signature (Key) -> LazyMethodModel
		private SortedList alternatives;  // Method alternatives

		public LazyObjectModel (Type type)
		{
			this.type  = type;
			this.fields = new ArrayList();
			this.methods = new SortedList();
			this.alternatives = new SortedList();

			ParseClass(type);
		}

		public Type Type {
			get { return type; }
		}

		public int FieldCount {
			get { return fields.Count; }
		}

		public LazyFieldModel GetField(int i) {
			return (LazyFieldModel) fields[i];
		}

		public int MethodCount {
			get { return methods.Count; }
		}

		public LazyMethodModel GetMethod(int i) {
			return (LazyMethodModel) methods.GetByIndex(i);
		}

		public LazyMethodModel GetMethod(string key) {
			return (LazyMethodModel) methods[key];
		}

		public int LazyMethodCount {
			get { return alternatives.Count; }
		}
		
		private ArrayList GetLazyMethod (int i) {
			return (ArrayList) alternatives.GetByIndex(i);
		}

		public int GetLazyMethodImplementationCount (int i) {
			return GetLazyMethod(i).Count;
		}

		public LazyMethodModel GetLazyMethodImplementation (int i, int j) {
			return (LazyMethodModel) GetLazyMethod(i)[j];
		}

		


		// Salida estndar

		public override string ToString () 
		{
			int i,j;
			string str = IsLazy()? "[Lazy] ":"";

			str += type;
			str += " ("+FieldCount+" fields, "+MethodCount+" methods)";

			str += "\n\nFields:";

			for (i=0; i<FieldCount; i++)
				str += "\n - " + GetField(i).ToString();

			str += "\n\nMethods:";

			for (i=0; i<MethodCount; i++)
				str += "\n - " + GetMethod(i).ToString();

			str += "\n\nLazy methods: " + LazyMethodCount;

			for (i=0; i<LazyMethodCount; i++) {

				str += "\n - Lazy method "+ (i+1);

				for (j=0; j<GetLazyMethodImplementationCount(i); j++)
					str += "\n   "+GetLazyMethodImplementation(i,j).ToString();
			}


			return str;
		}

		
		// Anlisis de la estructura de la clase

		private const BindingFlags fieldFlags = 
			  BindingFlags.Public 
			| BindingFlags.NonPublic
			| BindingFlags.DeclaredOnly
			| BindingFlags.Instance;

		private const BindingFlags methodFlags = 
			BindingFlags.Public 
			| BindingFlags.NonPublic 
			| BindingFlags.DeclaredOnly
			| BindingFlags.Instance;

		private void ParseClass (Type type) {

			// ILReader

			ModuleReader reader = new ModuleReader(type.Module, new AssemblyProvider());

			// Datos de la clase
			
			FieldInfo[]  fieldsInfo = type.GetFields ( fieldFlags );

			foreach (FieldInfo field in fieldsInfo) {
				fields.Add(new LazyFieldModel(this,field));
			}
			
			// Mtodos [pblicos] de la clase

			MethodInfo[]    methodsInfo = type.GetMethods( methodFlags );
			
			foreach (MethodInfo method in methodsInfo) {
				methods.Add(LazyMethodModel.MethodKey(method), new LazyMethodModel(this,method,reader));
			}

			ConstructorInfo[] constructorsInfo = type.GetConstructors( methodFlags );

			foreach (ConstructorInfo constructor in constructorsInfo) {
				methods.Add(LazyMethodModel.MethodKey(constructor), new LazyMethodModel(this,constructor,reader));
			}

			// Alternatives

			foreach (LazyMethodModel method in methods.Values)  {

				method.ComputeDefUseClosure();

				if (method.IsAlternativeImplementation()) {
					BuildAlternatives(method);
				}
			}
		}


		private void BuildAlternatives (LazyMethodModel method) {

			string    key  = method.AlternativeKey();
			ArrayList list = (ArrayList) alternatives[key];

			if (list==null) {
				list = new ArrayList();
				list.Add( methods[key] );
			}

			list.Add(method);

			alternatives[key] = list;
		}

		

		// IsLazy

		public bool IsLazy () 
		{
			object[] attributes = type.GetCustomAttributes(typeof(LazyAttribute),false);

			return (attributes.Length>0);
		}

		// Method alternative to be invoked
		// (i.e. the one which uses more instance variables)

		public int LazyMethodIndex (object obj, int lazy) {

			int best = 0;
			int weight = 0;

			LazyMethodModel method;

			for (int i=0; i<GetLazyMethodImplementationCount(lazy); i++) {

				method = GetLazyMethodImplementation(lazy,i);

				if (method.CheckObject(obj)) {

					if (method.UseCount()>weight) {
						weight = method.UseCount();
						best = i;					
					}
				}
			}

			return best;
		}

		public LazyMethodModel LazyMethod (object obj, int lazy) {

			return GetLazyMethodImplementation(lazy, LazyMethodIndex(obj, lazy) );
		}

		// Method invocation

		public object InvokeMethod (object obj, int alternative, object[] parameters) 
		{
			return LazyMethod(obj,alternative).Method.Invoke(obj,parameters);
		}

		public object InvokeMethodImplementation (object obj, int lazy, int alternative, object[] parameters) {
			return GetLazyMethodImplementation(lazy,alternative).Method.Invoke(obj,parameters);
		}



	}
}
