using System;
using System.Reflection;
using System.Reflection.Emit;

namespace EmitTest
{
	/// <summary>
	/// Ejemplo de creacin de objetos cuyo comportamiento cambia dinmicamente
	/// </summary>
	class EmitExample
	{
	
		[STAThread]
		static void Main(string[] args)
		{
				// Standard object

				BaseClass baseObject = new BaseClass(0,0,0,"Base");

				baseObject.Set(1);
				Console.WriteLine(baseObject.ToString());

				// Standard subclass object

				BaseClass subObject = new SubClass(0,0,0);

				subObject.Set(1);
				Console.WriteLine(subObject.ToString());

				((SubClass)subObject).TestReflection();
				Console.WriteLine(subObject.ToString());

				// Derived object (by reflection)

				Type derivedType = CreateDynamicType();
				object[] constructorArgs = new Object[] { 0, 0, 0, "Reflection"};

				BaseClass derivedObject = (BaseClass) Activator.CreateInstance(derivedType,constructorArgs);

				derivedObject.Set(1);
				Console.WriteLine(derivedObject.ToString());
			    derivedObject.Set(2);
			    Console.WriteLine(derivedObject.ToString());
				derivedObject.Set(3);
				Console.WriteLine(derivedObject.ToString());

		}


		private static Type CreateDynamicType () 
		{
			TypeBuilder myTypeBuilder = CreateTypeBuilder();

			FieldInfo myMethodName = DefineData (myTypeBuilder);

			DefineConstructor(myTypeBuilder, myMethodName);
			DefineMethod(myTypeBuilder, myMethodName);
						
			return myTypeBuilder.CreateType();
		}
		
		
		private static TypeBuilder CreateTypeBuilder ()
		{
			AppDomain domain = AppDomain.CurrentDomain;
			Type   baseType = typeof(BaseClass);
			string newTypeName = "DerivedClass";

			// Create an assembly.

			AssemblyName myAssemblyName = new AssemblyName();

			myAssemblyName.Name = "DynamicAssembly";

			AssemblyBuilder myAssembly =
				domain.DefineDynamicAssembly(myAssemblyName,AssemblyBuilderAccess.Run);
			
			// dynamic module in Dynamic Assembly.

			ModuleBuilder myModuleBuilder=myAssembly.DefineDynamicModule("LazyModule");		

			// public class in the assembly

			return myModuleBuilder.DefineType(newTypeName,TypeAttributes.Public,baseType);
		}
		
		
		// DATOS (private string field: invoked method name)

		private static FieldBuilder DefineData (TypeBuilder myTypeBuilder) {
			
			string alternativeField = "MyAlternative";
			
			return myTypeBuilder.DefineField
				(alternativeField, typeof(string), FieldAttributes.Private);
		}
			
			
		// CONSTRUCTOR

		private static void DefineConstructor (TypeBuilder myTypeBuilder, FieldInfo myMethodName) 
		{
			Type   baseType = typeof(BaseClass);

			Type[] constructorArgs = { typeof(int), typeof(int), typeof(int), typeof(string) };
			ConstructorBuilder constructor = myTypeBuilder.DefineConstructor(
				MethodAttributes.Public, CallingConventions.Standard, constructorArgs);

			ILGenerator constructorIL = constructor.GetILGenerator();

			// Llamada al constructor base

			constructorIL.Emit(OpCodes.Ldarg_0);   // this
			constructorIL.Emit(OpCodes.Ldarg_1);   // 1st param
			constructorIL.Emit(OpCodes.Ldarg_2);   // 2nd param
			constructorIL.Emit(OpCodes.Ldarg_3);   // 3rd param
			constructorIL.Emit(OpCodes.Ldarg_S,4); // 4th param
			ConstructorInfo superConstructor = baseType.GetConstructor(constructorArgs);
			constructorIL.Emit(OpCodes.Call, superConstructor);

			// Inicializacin de los mtodos activos


			constructorIL.Emit(OpCodes.Ldarg_0); // this
			constructorIL.Emit(OpCodes.Ldstr, "Set");		// Set, SetX, SetY
			constructorIL.Emit(OpCodes.Stfld, myMethodName);
			
			constructorIL.Emit(OpCodes.Ret);
		}


		// MTODO

		static Type   baseType = typeof(BaseClass);
		static string methodName = "Set";
		static string alternativeMethodName = "SetY";
		static Type[] methodArgs = { typeof(int) };
		static Type   returnType = null; // typeof(String);

		private static void DefineMethod (TypeBuilder myTypeBuilder, FieldInfo myMethodName) 
		{
			MethodAttributes attributes = MethodAttributes.Public | MethodAttributes.Virtual;

			MethodBuilder myMethodBuilder= myTypeBuilder.DefineMethod
				(methodName, attributes, returnType, methodArgs);

			ILGenerator methodIL = myMethodBuilder.GetILGenerator();

			// Method call
			// -----------

			ConditionalCall(methodIL,myMethodName);

			// - DEFAULT
			// Default(methodIL, alternativeMethodName, methodArgs);

			// - DYNAMIC
			//DynamicCall(methodIL, myMethodName);
		

			// Method update
			// -------------

			MethodUpdate(methodIL,myMethodName);

			// Optional operations...

			SampleOps(methodIL);

			// return

			methodIL.Emit(OpCodes.Ret);
		}

		
		private static void SampleOps (ILGenerator methodIL) 
		{
  			BindingFlags fieldFlags = BindingFlags.Public 
				| BindingFlags.NonPublic 
				| BindingFlags.DeclaredOnly
				| BindingFlags.Instance;

			// z = z + y

			FieldInfo srcField = baseType.GetField("y",fieldFlags);
			FieldInfo dstField = baseType.GetField("z",fieldFlags);

			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldfld, srcField);		
			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldfld, dstField);
			methodIL.Emit(OpCodes.Add);
			methodIL.Emit(OpCodes.Stfld, dstField);
		}

	
		private static void MethodUpdate (ILGenerator methodIL, FieldInfo myMethodName) {
			methodIL.Emit(OpCodes.Ldarg_0);
			methodIL.Emit(OpCodes.Ldstr, alternativeMethodName);
			methodIL.Emit(OpCodes.Stfld, myMethodName);
		}
		
		
		private static void DefaultCall (ILGenerator methodIL, string methodName, Type[] methodArgs) 
		{
			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldarg_1); // 1st param
/*
			if (reqdParams.Length != 0) {
				for (int c=1; c<=reqdParams.Length; c++) {
					outputIL.Emit(OpCodes.Ldarg_S,c);
				}
			}
*/
			MethodInfo alternativeMethod = baseType.GetMethod(methodName, methodArgs);
			methodIL.Emit(OpCodes.Call, alternativeMethod);
		}
		

		private static void DynamicCall (ILGenerator methodIL, FieldInfo myMethodName) 
		{
			// Local variables (method parameters)

			LocalBuilder parameters = methodIL.DeclareLocal(typeof(object[]));
			LocalBuilder method = methodIL.DeclareLocal(typeof(MethodInfo));

			// Ancillary routine (Console.Writeline)
			// MethodInfo writeMI = typeof(Console).GetMethod(
			// 	"WriteLine", new Type[] {typeof(object)});

			// parameters = new object[X]
			methodIL.Emit(OpCodes.Ldc_I4, methodArgs.Length);
			methodIL.Emit(OpCodes.Newarr, typeof(object));
			methodIL.Emit(OpCodes.Stloc, parameters);

			// parameters[0] = args
			methodIL.Emit(OpCodes.Ldloc, parameters);
			methodIL.Emit(OpCodes.Ldc_I4, 0);  // position 0
			methodIL.Emit(OpCodes.Ldarg_1);    // 1st parameter
			methodIL.Emit(OpCodes.Box, typeof(int));   // valor -> referencia (OJO!)
			methodIL.Emit(OpCodes.Stelem_Ref); // asignacin


			// this.GetType().GetMethod(X)
			// ---------------------------
			
			// this.GetType();
			methodIL.Emit(OpCodes.Ldarg_0);  // this
			MethodInfo getType = typeof(object).GetMethod("GetType");
			methodIL.Emit(OpCodes.Call, getType);

			// method name (private field)
			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldfld, myMethodName);
			// FIXED: methodIL.Emit(OpCodes.Ldstr, "Set");
						
			// method info (type, method name)
			Type[] getMethodParameters = new Type[] { typeof(string) };
			MethodInfo getMethod = typeof(Type).GetMethod("GetMethod", getMethodParameters);	
			methodIL.Emit(OpCodes.Callvirt, getMethod);
			
			methodIL.Emit(OpCodes.Stloc, method);
				
			// methodIL.Emit(OpCodes.Ldloc, method);    
			// methodIL.EmitCall(OpCodes.Call, writeMI, null);

			// Invoke
			// ------
			// Llamada recursiva (this.set) -> stack overflow (base.set)
			
			// parameters	
			methodIL.Emit(OpCodes.Ldloc, method);
			methodIL.Emit(OpCodes.Ldarg_0); // this 
			methodIL.Emit(OpCodes.Ldloc,parameters); // object[]
			// method invocation
			Type[] invokeParameters = new Type[] { typeof(object), typeof(object[]) };
			MethodInfo invoke = typeof(MethodInfo).GetMethod("Invoke",invokeParameters);
			methodIL.Emit(OpCodes.Callvirt, invoke);

			if (returnType==null)
				methodIL.Emit(OpCodes.Pop);  // void ...		
		}


		private static void ConditionalCall (ILGenerator methodIL, FieldInfo myMethodName) {
			Type   baseType = typeof(BaseClass);
			Label labelDynamic = methodIL.DefineLabel();
			Label labelEnd = methodIL.DefineLabel();

			// method name (private field)
			methodIL.Emit(OpCodes.Ldarg_0); // this
			methodIL.Emit(OpCodes.Ldfld, myMethodName);

			methodIL.Emit(OpCodes.Ldstr, methodName);

			Type[] strEqualsParameters = new Type[] { typeof(string), typeof(string) };
			MethodInfo strEquals = typeof(String).GetMethod("op_Equality", strEqualsParameters);
			methodIL.Emit(OpCodes.Call, strEquals);

			methodIL.Emit(OpCodes.Brfalse_S,labelDynamic);

			DefaultCall(methodIL, methodName, methodArgs);
			methodIL.Emit(OpCodes.Br,labelEnd);
			
			methodIL.MarkLabel(labelDynamic);
			DynamicCall(methodIL,myMethodName);

			methodIL.MarkLabel(labelEnd);
		}

	
	}
}
